#include <iostream>
#include <unistd.h>
#include <assert.h>

#include "strmanip.h"
#include "immsdb.h"

using std::endl;
using std::cerr; 

#define MAX(x,y) (x > y ? x : y)
#define MIN(x,y) (x < y ? x : y)

ImmsDb::ImmsDb()
    : SqlDb(string(getenv("HOME")).append("/.imms/imms.db"))
{
    sql_schema_upgrade();
    sql_create_tables();
}

ImmsDb::~ImmsDb()
{
    expire_recent("");
}

void ImmsDb::sql_create_tables()
{
    run_query(
            "CREATE TABLE 'Library' ("
                "'uid' INTEGER NOT NULL, "
                "'sid' INTEGER DEFAULT '-1', "
                "'path' VARCHAR(4096) UNIQUE NOT NULL, "
                "'modtime' TIMESTAMP NOT NULL, "
                "'checksum' VARCHAR(34) NOT NULL);");

    run_query(
            "CREATE TABLE 'Rating' ("
                "'uid' INTEGER UNIQUE NOT NULL, "
                "'rating' INTEGER NOT NULL);");

    run_query(
            "CREATE TABLE 'Info' ("
                "'sid' INTEGER UNIQUE NOT NULL," 
                "'artist' VARCHAR(255) NOT NULL, "
                "'title' VARCHAR(255) NOT NULL);");

    run_query(
            "CREATE TABLE 'Last' ("
                "'sid' INTEGER UNIQUE NOT NULL, " 
                "'last' TIMESTAMP);");

    run_query(
            "CREATE TABLE 'Correlations' ("
                "'origin' INTEGER NOT NULL, "
                "'destination' INTEGER NOT NULL, "
                "'weight' INTEGER DEFAULT '0');");

    run_query(
            "CREATE TEMPORARY TABLE 'Recent' ("
                "'sid' INTEGER NOT NULL, "
                "'weight' INTEGER NOT NULL, "
                "'time' TIMESTAMP);");
}

void ImmsDb::add_recent(int weight)
{
    expire_recent("WHERE time < '" + itos(time(0) - CORRELATION_TIME) + "'");

    if (sid != -1)
        run_query("INSERT INTO 'Recent' "
                "VALUES ('" + itos(sid) + "', '"
                + itos(weight) + "', '"
                + itos(time(0)) + "')");
}

void ImmsDb::expire_recent(const string &where_clause)
{
    while (1)
    {
        select_query(
                "SELECT sid, weight FROM 'Recent' "
                + where_clause + " LIMIT 1;");

        if (!nrow)
            return;

        int from = atoi(resultp[ncol]);
        int from_weight = atoi(resultp[ncol + 1]);

        run_query("DELETE FROM 'Recent' WHERE sid = '" + itos(from) + "';");

        // FIXME: Consider using the query callbacks for this
        select_query("SELECT sid, weight FROM 'Recent';");

        for (; nrow > 0; --nrow)
        {
            int to = atoi(resultp[nrow * ncol]);
            int to_weight = atoi(resultp[nrow * ncol + 1]);

            if (to == from)
                continue;

            if (from_weight < 0 && to_weight < 0)
                continue;

            float weight = sqrt(abs(from_weight*to_weight));

            if (from_weight < 0 || to_weight < 0)
                weight = -weight;

            run_query("INSERT INTO 'Correlations' ('origin', 'destination') "
                    "VALUES ('" + itos(MIN(from, to)) + "', "
                    "'" + itos(MAX(from, to)) + "');");

#ifdef DEBUG
            cerr << " >> Updating link from " << MIN(from, to) << " to "
                << MAX(from, to) << " by " << weight << endl;
#endif

            run_query(
                    "UPDATE 'Correlations' SET weight = weight + '" +
                    itos(weight) + "' "
                    "WHERE origin = '" + itos(MIN(from, to)) + "' "
                    "AND destination = '" + itos(MAX(from, to)) + "';");
        }
    }
}

int ImmsDb::correlate(int from)
{
    if (sid == -1)
        return 0;
    
    select_query(
            "SELECT weight FROM 'Correlations' "
            "WHERE origin = '" + itos(MIN(from, sid)) + "' "
                "AND destination = '" + itos(MAX(from, sid)) + "';");

    return nrow && resultp[1] ? atoi(resultp[1]) : 0;
}

int ImmsDb::identify(const string &path, time_t modtime)
{
    title = artist = "";

    select_query(
            "SELECT uid, sid, modtime FROM 'Library' "
            "WHERE path = '" + escape_string(path) + "';");

    sid = uid = -1;

    if (nrow)
    {
        uid = atoi(resultp[ncol]);
        sid = atoi(resultp[ncol + 1]);
        time_t last_modtime = atol(resultp[ncol + 2]);

        if (modtime == last_modtime)
            return uid;
    }

    return -1;
}

int ImmsDb::identify(const string &path, time_t modtime, const string &checksum)
{
    // old path but modtime has changed - update checksum
    if (nrow)
    {
        run_query(
                "UPDATE 'Library' SET "
                    "modtime = '" + itos(modtime) + "', "
                    "checksum = '" + checksum + "' "
                "WHERE path = '" + escape_string(path) + "';");
        return uid;
    }

    // moved or new file and path needs updating

    sid = uid = -1;

    select_query(
            "SELECT uid, sid, path FROM 'Library' "
            "WHERE checksum = '" + checksum + "';");

    if (nrow)
    {
        // Check if any of the old paths no longer exist 
        // (aka file was moved) so that we can reuse their sid
        for (int col = ncol; nrow--; col += ncol)
        {
            uid = atoi(resultp[col]);
            sid = atoi(resultp[col + 1]);

            if (access(resultp[col + 2], F_OK))
            {
                run_query(
                        "UPDATE 'Library' SET "
                            "sid = '-1', "
                            "path = '" + escape_string(path) + "', "
                            "modtime = '" + itos(modtime) + "' "
                        "WHERE path = '" + resultp[col + 2] + "';");

                return uid;
            }
        }
    }
    else
    {
        // figure out what the next uid should be
        select_query("SELECT max(uid) FROM Library;");
        uid = resultp[1] ? atoi(resultp[1]) + 1 : 1;
    }

    // new file - insert into the database
    run_query(
            "INSERT INTO 'Library' ('uid', 'path', 'modtime', 'checksum') "
            "VALUES ('" +
                itos(uid) + "', '" +
                escape_string(path) + "', '" +
                itos(modtime) + "', '" +
                checksum + "'"
            ");");

    return uid;
}

bool ImmsDb::check_artist(string &artist)
{
    select_query(
            "SELECT artist FROM 'Info' "
            "WHERE artist LIKE '" + artist + "' LIMIT 1;");

    if (nrow && resultp[1])
    {
        artist = resultp[1];
        return true;
    }
    return false;
}

bool ImmsDb::check_title(string &title)
{
    select_query(
            "SELECT title FROM 'Info' "
            "WHERE artist = '" + artist + "' "
                "AND title LIKE '" + title + "' LIMIT 1;");

    if (nrow && resultp[1])
    {
        title = resultp[1];
        return true;
    }
    return false;
}

pair<string, string> ImmsDb::get_info()
{
    if (sid == -1)
        return StringPair("", "");

    select_query(
            "SELECT title, artist FROM 'Info' "
            "WHERE sid = '" + itos(sid) + "';");

    artist = nrow ? resultp[3] : "";
    title = nrow ? resultp[2] : "";
#ifdef DEBUG_1
    if (title != "" && artist != "")
    {
        cerr << " >> from cache: " << endl;
        cerr << " >>     artist  = '" << artist << "'" << endl;
        cerr << " >>     title   = '" << title << "'" << endl;
    }
#endif
    return StringPair(artist, title);
}

time_t ImmsDb::get_last()
{
    if (sid == -1)
        return 0;

    select_query(
            "SELECT last FROM 'Last' "
            "WHERE sid = '" + itos(sid) + "';");

    return nrow && resultp[1] ? atol(resultp[1]) : 0;
}

int ImmsDb::get_rating()
{
    if (uid == -1)
        return -1;

    select_query(
            "SELECT rating FROM 'Rating' "
            "WHERE uid = '" + itos(uid) + "';");

    return nrow ? atoi(resultp[1]) : -1;
}

void ImmsDb::set_artist(const string &_artist)
{
    artist = _artist;
}

void ImmsDb::set_title(const string &_title)
{
    if (uid == -1)
        return;

    title = _title;

    // clean up the old info
    if (sid != -1)
    {
        string where_clause = "WHERE sid = '" + itos(sid) + "'";
        run_query("DELETE FROM 'Info' " + where_clause + ";");
        run_query("DELETE FROM 'Last' " + where_clause + ";");
    }

    select_query(
            "SELECT sid FROM 'Info' "
            "WHERE artist = '" + artist + "' AND title = '" + title + "';");

    register_new_sid(nrow && resultp[1] ? atoi(resultp[1]) : sid);

    run_query(
            "INSERT INTO 'Info' ('sid', 'artist', 'title') "
            "VALUES ('" + itos(sid) + "', '" + artist + "', '" + title + "');");

#ifdef DEBUG
    cerr << " >> cached as: " << endl;
    cerr << " >>     artist  = '" << artist << "'" << endl;
    cerr << " >>     title   = '" << title << "'" << endl;
#endif
}

void ImmsDb::register_new_sid(int new_sid)
{
    if (new_sid == -1)
    {
        select_query("SELECT max(sid) FROM Library;");
        new_sid = resultp[1] ? atoi(resultp[1]) + 1 : 0;
    }
    else
    {
        run_query(
                "UPDATE 'Correlations' SET origin = '" + itos(new_sid) + "' "
                "WHERE origin = '" + itos(sid) + "';");

        run_query(
                "UPDATE 'Correlations' SET destination = '" + itos(new_sid) + 
                "' WHERE destination = '" + itos(sid) + "';");
    }

    sid = new_sid;

    run_query(
            "UPDATE 'Library' SET sid = '" + itos(sid) + "' "
            "WHERE uid = '" + itos(uid) + "';");
}

void ImmsDb::set_last(time_t last)
{
    if (uid == -1)
        return;

    if (sid == -1)
        register_new_sid();

    run_query(
            "INSERT OR REPLACE INTO 'Last' ('sid', 'last') "
            "VALUES ('" + itos(sid) + "', '" +  itos(last) + "');");
}

void ImmsDb::set_rating(int rating)
{
    if (uid == -1)
        return;

    run_query(
            "INSERT OR REPLACE INTO 'Rating' ('uid', 'rating') "
            "VALUES ('" + itos(uid) + "', '" +  itos(rating) + "');");
}

void ImmsDb::sql_schema_upgrade()
{
    select_query("SELECT version FROM 'Schema' WHERE description ='latest';");

    if (nrow && resultp[1] && atoi(resultp[1]) > SCHEMA_VERSION)
    {
        cerr << "IMMS: Newer database schema detected." << endl;
        cerr << "IMMS: Please update IMMS!" << endl;
        close_database();
        return;
    }

    if (nrow && resultp[1] && atoi(resultp[1]) == SCHEMA_VERSION)
        return;

    cerr << "IMMS: Outdated database schema detected." << endl;
    cerr << "IMMS: Attempting to update." << endl;

    run_query("DROP TABLE Info;");
    run_query("DROP TABLE Last;");
    run_query("DROP TABLE UnknownLast;");

    // Backup the existing tables
    run_query(
            "CREATE TEMPORARY TABLE Library_backup ("
                "'sid' INTEGER NOT NULL, "
                "'path' VARCHAR(4096) UNIQUE NOT NULL, "
                "'modtime' TIMESTAMP NOT NULL, "
                "'checksum' VARCHAR(34) NOT NULL);");

    run_query("INSERT INTO Library_backup SELECT * FROM Library;");
    run_query("DROP TABLE Library;");

    run_query(
            "CREATE TEMPORARY TABLE Rating_backup ( "
                "'uid' INTEGER UNIQUE NOT NULL, "
                "'rating' INTEGER NOT NULL);");

    run_query("INSERT INTO Rating_backup SELECT * FROM Rating;");
    run_query("DROP TABLE Rating;");

    // Create new tables
    sql_create_tables();

    // Copy the data into new tables, and drop the backups
    run_query(
            "INSERT INTO Library (uid, path, modtime, checksum) "
            "SELECT * FROM Library_backup;");

    run_query("DROP TABLE Library_backup;");

    run_query("INSERT INTO Rating SELECT * FROM Rating_backup;");

    run_query("DROP TABLE Rating_backup;");

    // Mark the new version
    run_query(
            "CREATE TABLE 'Schema' ("
                "'description' VARCHAR(10) UNIQUE NOT NULL, "
                "'version' INTEGER NOT NULL);");

    run_query(
            "INSERT OR REPLACE INTO 'Schema' ('description', 'version') "
            "VALUES ('latest', '" +  itos(SCHEMA_VERSION) + "');");
}
